#*****************************************************************************
#
#    Sage extension class "MTX", based on the Aachen Meataxe
#
#    Copyright (C) 2009 Simon A. King <simon.king@uni-jena.de>
#
#  Distributed under the terms of the GNU General Public License (GPL),
#  version 2 or later (at your choice)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  http://www.gnu.org/licenses/
#*****************************************************************************

include "interrupt.pxi"

#
# First, something from Python.h and string.h
#

cdef extern from "Python.h":
    object PyString_FromStringAndSize(char *s, Py_ssize_t len)
    char* PyString_AsString(object string)

#
# Next, import SOME types from meataxe.h
# (most types are not needed, but listed here 
# in the comments, for completeness)
#


cdef extern from "meataxe.h":
    # general ctype emulations
    ctypedef int size_t
    ctypedef unsigned long Ulong
    ctypedef unsigned short Ushort
    ctypedef unsigned char Uchar
    ctypedef struct FILE
    
    # assume small fields (default) 
    ctypedef unsigned char FEL
    ctypedef FEL *PTR
    cdef extern FEL tmult[256][256]
    cdef extern FEL tadd[256][256]
    cdef extern FEL taddinv[256], tmultinv[256]
    cdef extern FEL tinsert[8][256], textract[8][256]

    # pivots
    ctypedef struct piv_t:
        long i,m

    #
    # cdef extern long zfl, zchar 	#/* Field order and characteristic */
    # cdef extern FEL zgen		#/* Generator */
    # cdef extern long znoc		#/* Number of columns for row ops */
    cdef extern size_t zrowsize 	#/* Row size in memory */
    cdef extern size_t zrowsize_io	#/* Row size in files */
    # cdef extern char zzzversion[]
    # cdef extern char zzz_cc[]
        
#########################################
# function prototypes
    # FEL zadd(FEL a,FEL b)
    # FEL zneg(FEL a)
    # FEL zmul(FEL a,FEL b)
    # FEL zinv(FEL a)
    #
    # FEL zembed(FEL a, long subfield)	#/* Embed from subfield */
    # FEL zrestrict(FEL a, long subfield)	#/* Restrict to subfield */
    FEL zitof(long l)
    long zftoi(FEL f)

    int zsetlen(long ncols)		#/* Set row size */
    int zsetfield(long field)
    void zinsert(PTR row, long col, FEL mark)
    FEL zextract(PTR row, long col)
    void zinsert_step(PTR *pos, int *idx, FEL mark)
    FEL zextract_step(PTR *pos, int *idx)

    piv_t _zfindpiv_(PTR row)
    # PTR zextractcol(PTR mat, long nor, long col, PTR result)
    # PTR zaddrow(PTR dest, PTR src)
    # PTR zsubrow(PTR dest, PTR src)
    # PTR znegrow(PTR row)
    PTR zmulrow(PTR row, FEL mark)
    # PTR zaddmulrow(PTR dest, PTR src, FEL f)
    # PTR zmaprow(PTR row, PTR matrix, long nor, PTR result)
    # PTR zpermrow(PTR row, long *perm, PTR result)

    void zmoverow(PTR dest, PTR src)
    void zcpyrow(PTR dest, PTR src, long sz)
    # void zswaprow(PTR dest, PTR src)
    # int zcmprow(PTR dest, PTR src)
    size_t zsize(long nrows)
    # PTR zalloc(long NROWS)
    void zadvance(PTR *ptr, long nrows)
    void zsteprow(PTR *ptr)
    void zbackrow(PTR *ptr)
    PTR ptrPlus(PTR base, long offset)
    # int mtxinit()

    # skip gap-conversion
    # bitstring operations
    # ctypedef struct bitstring_t:
    #     Ushort id	#/* always = T_BITS */
    #     char dummy[sizeof(long)-sizeof(Ushort)]
    #     unsigned char buf[1]
    # cdef extern size_t bs_size
    # void bs_print(bitstring_t *b)
    # char *bs_desc(bitstring_t *b)
    # void bs_setlen(int l)
    # void bs_reset(bitstring_t *x)
    # bitstring_t *bs_alloc()
    # bitstring_t *bs_dup(bitstring_t *s)
    # void bs_free(bitstring_t *b)
    # void bs_set(bitstring_t *b, int i)
    # void bs_clear(bitstring_t *b, int i)
    # int bs_test(bitstring_t *b, int i)
    # bitstring_t *bs_read(FILE *f)
    # int bs_write(FILE *f, bitstring_t *b)
    # void bs_and(bitstring_t *dest, bitstring_t *src)
    # void bs_minus(bitstring_t *dest, bitstring_t *src)
    # void bs_or(bitstring_t *dest, bitstring_t *src)
    # int bs_match(bitstring_t *x, bitstring_t *y)
    # int bs_issub(bitstring_t *a, bitstring_t *b)
    # int bs_cmp(bitstring_t *a, bitstring_t *b)
    # void bs_cpy(bitstring_t *a, bitstring_t *b)

    # file i/o
    # size_t zreadlong(FILE *f, long *buf, size_t n)
    # size_t zwritelong(FILE *f, long *buf, size_t n)
    # size_t zreadvec(FILE *f, PTR buf, size_t n)
    # size_t zwritevec(FILE *f, PTR buf, size_t n)
    # int zseek(FILE *f, long pos)
    # FILE *zreadhdr(char *name, long *fld, long *nor, long *noc)
    # FILE *zwritehdr(char *name, long fld, long nor, long noc)

    # Gauss elimination and related functions
    # int zmkpivot(PTR matrix, long nor, long *piv)
    # long zmkechelon(PTR matrix, long nor, long *piv)
    # int zcleanrow(PTR row, PTR matrix, long nor, long *piv)
    # int zcleanrow2(PTR row,PTR matrix,long nor,long *piv,PTR row2)
    # long znullsp(PTR matrix, long nor, long *piv, PTR nsp)
    # int zmatinv(PTR mat, PTR result)

    #  Projection on the quotient
    # int zquotinit(PTR subspace, long dim, long *piv)
    # int zquot(PTR space, long dim, PTR quot)
    # int zquotop(PTR matrix, PTR quot)

    # Seed vector generation
    # cdef extern PTR zpseed_vec		#/* The seed vector */
    # void zpseed_free()
    # int zpseed_init(long sdim, PTR sbasis)
    # long zpseed_make(long num)
    #long zpseed_next()

    # Command line parsing
    # ctypedef struct proginfo_t:
    #     char *name
    #     char *shortdesc
    #     char *rcsrev
    #     char **helptext
    # cdef extern char opt_char		#/* Current option */
    # cdef extern char opt_text[50]	#/* Option text */
    # cdef extern char *opt_text_ptr	#/* Current position in option text */
    # cdef extern int opt_ind		#/* Index in argv[] list */
    # cdef extern char MeatAxeBinDir[250]	#/* MeatAxe program directory */

    # void initargs(int argc, char **argv, proginfo_t *pi)
    # int zgetopt(char *pattern)
    # long getint()

    # Messages
    # cdef extern int mtxerrno
    # cdef extern int mtxerraction
    # void mtxerror(char *text)
    # void errexit(int code, char *text)
    # int errhandler(int errno, char *errfile, int errline)
    # int Message(FILE *f, char *fmt, ...)
    # int ErrorExit(char *fmt, ...)
    # cdef extern int msg_level

    # CPU time
    # void prtimes()

    # Random numbers
    # void RandInit(unsigned seed)
    # long int Random()

    # Miscellaneous
    # long gcd(long a, long b)
    # long lcm(long a, long b)

    # Matrices
    ctypedef struct matrix_t:
        Ushort id			#/* Always = T_MATRIX */
        long fl, nor, noc		#/* Field, #rows, #columns */
        PTR d			        #/* Pointer to data area */

    # void matprint(char *name, matrix_t *p)
    # char *matdesc(matrix_t *s)
    matrix_t *matalloc(long fl, long nor, long noc)
    matrix_t *MTXmatid(long fl, long nor)
    void matfree(matrix_t *m)

    long matcmp(matrix_t *m1, matrix_t *m2)
    matrix_t *matdup(matrix_t *src)
    # int  matmove(matrix_t *dest, matrix_t *src)
    matrix_t *_matextract(matrix_t *src, long first, long n)
    matrix_t *matread(FILE *f)
    int matwrite(FILE *f, matrix_t *mat)
    matrix_t *matload(char *fn)
    int matsave(matrix_t *mat, char *fn)

    matrix_t *matadd(matrix_t *dest, matrix_t *src)
    matrix_t *matmul(matrix_t *dest, matrix_t *src)
    int matmul_result(matrix_t *dest, matrix_t *src, matrix_t *res)
    matrix_t *matmulF(matrix_t *src, FEL mark)
    matrix_t *matsub(matrix_t *dest, matrix_t *src)
    matrix_t *matneg(matrix_t *src)
    matrix_t *mattr(matrix_t *src)
    matrix_t *matinv(matrix_t *src)
    # int chkechelon(matrix_t *mat)
    matrix_t *matpower(matrix_t *mat, long n)
    long matorder(matrix_t *mat)
    
    # Higher-level gauss functions
    matrix_t *echelon(matrix_t *mat)
    long nullity(matrix_t *mat)
    matrix_t *nullspace(matrix_t *mat)

    # ycomp.c
    # int spccomp(matrix_t *m1, matrix_t *m2, long n)
    # int spcequal(matrix_t *m1, matrix_t *m2)
    # int spccontains(matrix_t *m1, matrix_t *m2)

    # yspin.c
    # matrix_t *matspin(matrix_t *seed, int ngen, matrix_t *gen[])
    # matrix_t *matspin_f(matrix_t *seed, int ngen, matrix_t *gen[], long *sdim)

    # Spin-up, split, and standard basis
    # long zspinup(PTR space, long nseed, long *piv, int ngen, PTR gen[],	int gentype)
    # int zsbasis(PTR seed, long nseed, int ngen, PTR gen[], PTR space, long *piv, PTR basis)
    # cdef extern long *split_pivot

    # ... various options in meataxe.h
    # int spinup(matrix_t *seed,int ngen,matrix_t *gen[],int options, matrix_t **subspace)
    # int split(matrix_t *subspace, int ngen, matrix_t *gen[], matrix_t **sub, matrix_t **quot)
    # matrix_t *quotproj(matrix_t *subspace, matrix_t *vectors)
    # matrix_t *sbasis(matrix_t *seed, int ngen, matrix_t **gen)
    # int chbasis(matrix_t *basis, int ngen, matrix_t **gen,matrix_t **newgen)

    # Polynomials
    # ctypedef struct poly_t:
    #     Ushort id			#/* Always = T_POLY */
    #     long fl
    #     long deg
    #     size_t size  #/* deg <= size-1 */
    #     FEL *buf

    # char *poldesc(poly_t *s)
    # poly_t *polalloc(long fl, long degree)
    # void polfree(poly_t *x)
    # poly_t *poldup(poly_t *x)
    # poly_t *poladd(poly_t *dest, poly_t *src)
    # poly_t *polmul(poly_t *dest, poly_t *src)
    # poly_t *poldivmod(poly_t *a, poly_t *b)
    # poly_t *polmod(poly_t *a, poly_t *b)
    # void polprint(char *name, poly_t *p)
    # poly_t *polread(FILE *f)
    # int polwrite(FILE *f, poly_t *p)
    # int polcmp(poly_t *a, poly_t *b)
    # poly_t *polgcd(poly_t *a, poly_t *b)
    # poly_t *polderive(poly_t *p)
    # int vec2pol(PTR vec, poly_t *pol)
    # int pol2vec(poly_t *pol, PTR vec)
    # int polpack(poly_t *pol, PTR vec)
    # poly_t *polshiftmod(poly_t *p, long n, poly_t *q)

    # The word generator
    # ctypedef struct wgdata_t:
    #     int ngen
    #     size_t len, max
    #     matrix_t **b
    # wgdata_t *WGInit(int ngen, matrix_t *gen[])
    # int WGFree(wgdata_t *b)
    # matrix_t *MakeWord(wgdata_t *b, long n)
    # void makefp(wgdata_t *b, long fp[])
    # char *SymbolicName(wgdata_t *b, long n)

    # Permutations
    # ctypedef struct perm_t:
    #     Ushort id			#/* Always = T_PERM */
    #     long deg
    #     PTR d
    # char *permdesc(perm_t *s)
    # void permprint(char *name, perm_t *x)
    # perm_t *permalloc(long deg)
    # void permfree(perm_t *m)
    # perm_t *permdup(perm_t *src)
    # perm_t *permmove(perm_t *dest, perm_t *src)
    # perm_t *permread(FILE *f)
    # int permwrite(FILE *f, perm_t *perm)
    # perm_t *permload(char *fn)
    # int permsave(perm_t *perm, char *fn)
    # perm_t *permmul(perm_t *dest, perm_t *src)
    # long permorder(perm_t *perm)
    # perm_t *permpower(perm_t *p, long n)

    # Insertion
    # matrix_t *matinsert(matrix_t *mat, poly_t *pol)
    # matrix_t *matinsert_(matrix_t *mat, poly_t *pol)

    # Factored polynomials
    # ctypedef struct fpoly_t:
    #     Ushort id			#/* Always T_FPOLY */
    #     size_t len, max
    #     poly_t **p			#/* Irreducible factors */
    #     long *e			#/* Multiplicities */
    # char *fpoldesc(fpoly_t *p)
    # void fpolprint(char *name, fpoly_t *p)
    # fpoly_t *fpolalloc()
    # fpoly_t *fpoldup(fpoly_t *s)
    # void fpolfree(fpoly_t *p)
    # fpoly_t *fpolread(FILE *f)
    # int fpolwrite(FILE *f, fpoly_t *p)
    # fpoly_t *fpolmulp(fpoly_t *dest, poly_t *src, long pwr)
    # fpoly_t *fpolmul(fpoly_t *dest, fpoly_t *src)

    # Characteristic and minimal polynomials
    # cdef extern long CharPolSeed
    # poly_t *charpolfactor(matrix_t *mat)
    # fpoly_t *charpol(matrix_t *mat)
    # poly_t *minpolfactor(matrix_t *mat)
    # fpoly_t *minpol(matrix_t *mat)

    # Berlekamp factorization
    # fpoly_t *factorization(poly_t *pol)

    # Sets
    # ctypedef struct set_t:
    #     Ushort id			#/* Always = T_SET */
    #     size_t len, max
    #     long *buf
    # int set_allocstrategy(size_t first,size_t blocksize)
    # char *set_desc(set_t *s)
    # void set_print(char *name, set_t *x)
    # set_t *set_alloc()
    # set_t *set_dup(set_t *s)
    # void set_free(set_t *x)
    # int set_insert(set_t *set, long elem)
    # int set_contains(set_t *set, long elem)
    # set_t *set_read(FILE *f)
    # int set_write(FILE *f, set_t *set)

    # Sequences
    # ctypedef struct sequence_t:
    #     Ushort id			#/* Always = T_SEQUENCE */
    #     size_t len, max
    #     void **buf
    # char *seq_desc(sequence_t *s)
    # void seq_print(char *name, sequence_t *s)
    # sequence_t *seq_alloc(size_t size)
    # sequence_t *seq_dup(sequence_t *s)
    # void seq_free(sequence_t *x)
    # int seq_insert(sequence_t *s, int pos, void *x)
    # int seq_remove(sequence_t *s, int pos)
    # sequence_t *seq_read(FILE *f)
    # int seq_write(FILE *f, sequence_t *x)

    # Integers and strings
    # ctypedef struct integer_t:
    #     Ushort id			#/* Always = T_INT */
    #     long l
    # integer_t *intalloc(long l)
    # integer_t *intdup(integer_t *s)
    # char *intdesc(integer_t *s)
    # void intprint(char *name, integer_t *x)
    # void intfree(integer_t *x)
    # integer_t *intread(FILE *f)
    # int intwrite(FILE *f, integer_t *x)
    # integer_t *intadd(integer_t *d, integer_t *s)
    # integer_t *intsub(integer_t *d, integer_t *s)
    # integer_t *intmul(integer_t *d, integer_t *s)
    # integer_t *intdiv(integer_t *d, integer_t *s)
    # integer_t *intneg(integer_t *x)

    # ctypedef struct string_t:
    #     Ushort id			#/* Always = T_STR */
    #     char *s
    # string_t *stringalloc(char *c)
    # string_t *stringdup(string_t *s)
    # char *stringdesc(string_t *s)
    # void stringprint(char *name, string_t *x)
    # void stringfree(string_t *x)
    # string_t *stringread(FILE *f)
    # int stringwrite(FILE *f, string_t *x)

    # MeatAxe object methods
    # void mtxfree(void *x)
    # void *mtxdup(void *x)
    # char *mtxgetname(void *x)
    # char *mtxgetdesc(void *x)
    # void *mtxread(FILE *f)
    # void mtxprint(void *x)
    # int mtxwrite(FILE *f, void *x)
    # void *mtxadd(void *l, void *r)
    # void *mtxsub(void *l, void *r)
    # void *mtxmul(void *l, void *r)
    # void *mtxdiv(void *l, void *r)
    # void *mtxpwr(void *l, void *r)
    # void *mtxneg(void *x)
    # void *mtxorder(void *x)

    # Profiling
    # cdef extern long ProfFileIO		#/* File I/O */
    # cdef extern long ProfNullSpace	#/* Matrix null space */
    # cdef extern long ProfSpinUp		#/* Spin-up */
    # cdef extern long ProfMakeWord	#/* Make word */
    # cdef extern long ProfMatInsert	#/* Matrix insertion */
    # cdef extern long ProfPolFactor	#/* Polynomial factorization */
    # cdef extern long ProfCharPol	#/* Char. and minimal polynomial */

    # Added by Simon King:
    cdef char* mat2str(matrix_t* m)           # Convert m->d into a char*
    cdef void str2mat(matrix_t* m, int l, char* x) # *COPY* the data from x to m->d
    cdef char* zrow2str(PTR d)       # Convert d into a char*
    cdef void zstr2row(PTR d, char* x) # *COPY* the data from x to d, assuming
                                       # that d has been assigned to at least
                                       # zrowsize_io byte.

# thats's all of meataxe.h !

cdef extern from "customize.h":
    void *copyrows(PTR dest, PTR src, size_t nor)

##############################################################

#
# In the .pxd-declaration file, we merely need to define the 
# cdef-parts of the extension class. Everything else is 
# in the .pyx-definition file.

cdef class MTX:
    cdef matrix_t *Data
    cdef object _mutable
    cpdef MTX _mulMTX_(self, MTX right)
    cpdef MTX _mulInt_(self, long right)
    cpdef int nrows(self)
    cpdef int ncols(self)
    cpdef int base(self)
    cpdef int characteristic(self)
    cpdef list _matlist_(self)
    cpdef MTX transpose(MTX self)
    cpdef int order(MTX self)
    #cpdef MTX stack(self, MTX N)
    cpdef MTX normalized(MTX self)
    cpdef MTX echelon(MTX self)
    cpdef MTX full_echelon(MTX self)
    cpdef list pivots(MTX self)
    cpdef int nullity(MTX self)
    cpdef MTX nullspace(MTX self)
    cpdef tuple lead(self)
